﻿/*
* V4L2 video, vbi and sound test capture and output (only mmap i/o method)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <vector>
#include <string>

#include <getopt.h>             /* getopt_long() */

#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <asm/types.h>          /* for videodev2.h */
#include <linux/videodev2.h>
#include <alsa/asoundlib.h>

#include "forward-v4l2-ioctl.h"

#define CLEAR(x) memset (&(x), 0, sizeof (x))
#define BUFF_NUMBERS 8

struct buffer {
	int index;
	void* start;
	size_t length;
};

/* video */
std::string dev_in = "/dev/video0";
std::string vbi_in = "/dev/vbi0";
std::string dev_out = "/dev/video2";
std::string vbi_out = "/dev/vbi2";
std::string audio_in = "hw:0,0";
std::string audio_out = "hw:0,2";

static int infd = -1;
static int outfd = -1;
static int invbifd = -1;
static int outvbifd = -1;

std::vector<buffer> inBuffers{ 0 };
std::vector<buffer> outBuffers{ 0 };
std::vector<buffer> inVBIBuffers{ 0 };
std::vector<buffer> outVBIBuffers{ 0 };

int intype = V4L2_BUF_TYPE_VIDEO_CAPTURE;
int outtype = V4L2_BUF_TYPE_VIDEO_OUTPUT;
int invbitype = V4L2_BUF_TYPE_VBI_CAPTURE;
int outvbitype = V4L2_BUF_TYPE_VBI_OUTPUT;

/* audio */
snd_pcm_t* pcm_in = nullptr, * pcm_out = nullptr;
snd_pcm_hw_params_t* hw_params_in = nullptr, * hw_params_out = nullptr;
snd_pcm_status_t* status_in = nullptr, * status_out = nullptr;
int32_t* buffer_in = nullptr, * buffer_out = nullptr;
snd_pcm_format_t format;
snd_pcm_uframes_t period = 1920;

static void errno_exit(const char* s)
{ 
	fprintf(stderr,"%s %d, %s\n", s, errno, strerror(errno));
	exit(EXIT_FAILURE);
}

static int xioctl(int fd, int  request, void* arg)
{
	int r;
	do r = ioctl(fd, request, arg);
	while (-1 == r && EINTR == errno);
	return r;
}

static void process_image(const void* p)
{
	fputc('.', stdout);
	fflush(stdout);
}

void sl_memcpy(void* buf_dest, void* buf_src, size_t len)
{

/* you can copy input frame to output */
	memcpy(buf_dest, buf_src, len);

/* or you can do something during coping */
	//using sometype = int64_t;
	//int k = len / sizeof(sometype);
	//sometype* bd = (sometype*)buf_dest;
	//sometype* bs = (sometype*)buf_src;
	//for (int i = 0; i < k; i++) bd[i] = bs[k - i - 1];
}

/* read from output and write to input */
static int rewrite_frame(void)
{
	struct v4l2_buffer inbuf;
	struct v4l2_buffer invbibuf;
	struct v4l2_buffer outbuf;
	struct v4l2_buffer outvbibuf;
	unsigned int i;

	/* read video */
	CLEAR(inbuf);
	inbuf.type = intype;
	inbuf.memory = V4L2_MEMORY_MMAP;
	if (-1 == xioctl(infd, VIDIOC_DQBUF, &inbuf)) {
		switch (errno) {
		case EAGAIN:
			return 0;
		case EIO:
			/* Could ignore EIO, see spec. */
			/* fall through */
		default:
			errno_exit("VIDIOC_DQBUF error");
		}
	}
	assert(inbuf.index < inBuffers.size());
	process_image(inBuffers[inbuf.index].start);
	if (-1 == xioctl(infd, VIDIOC_QBUF, &inbuf))
		errno_exit("VIDIOC_QBUF error");

	/* read VBI */
	CLEAR(invbibuf);
	invbibuf.type = invbitype;
	invbibuf.memory = V4L2_MEMORY_MMAP;
	if (-1 == xioctl(invbifd, VIDIOC_DQBUF, &invbibuf)) {
		switch (errno) {
		case EAGAIN:
			return 0;
		case EIO:
			/* Could ignore EIO, see spec. */
			/* fall through */
		default:
			errno_exit("VIDIOC_DQBUF error");
		}
	}
	assert(invbibuf.index < inVBIBuffers.size());
	if (-1 == xioctl(invbifd, VIDIOC_QBUF, &invbibuf))
		errno_exit("VIDIOC_QBUF error");

	snd_pcm_sframes_t size;
	snd_htimestamp_t ts;
	/* read audio */
	size_t lll = period * snd_pcm_format_width(format) / 8 * 8;
	size = snd_pcm_readi(pcm_in, buffer_in, period);
	if (size != period)
		fprintf(stderr,"Wrong number of frames readed: size = %d, period = %d\n", size, period);

	if (0 != snd_pcm_status(pcm_in, status_in))
		fprintf(stderr,"Cannot get input pcm status\n");

	snd_pcm_status_get_audio_htstamp(status_in, &ts);

	/* write audio */
	memcpy((void*)buffer_out, (void*)buffer_in, lll);

	size = snd_pcm_writei(pcm_out, buffer_out, period);
	if (size != period)
		fprintf(stderr,"Wrong number of frames written: size = %d, period = %d\n", size, period);

	if (0 != snd_pcm_status(pcm_out, status_out))
		fprintf(stderr,"Cannot get output pcm status\n");

	snd_pcm_status_get_audio_htstamp(status_out, &ts);

	/* obtain buffer for output video */
	CLEAR(outbuf);
	outbuf.type = outtype;
	if (-1 == xioctl(outfd, VIDIOC_DQBUF, &outbuf)) {
		switch (errno) {
		case EAGAIN:
			return 0;
		case EIO:
			/* Could ignore EIO, see spec. */
			/* fall through */
		default:
			errno_exit("VIDIOC_DQBUF error");
		}
	}
	assert(outbuf.index < outBuffers.size());
	/* copy image from input to output */
	memcpy((void*)outBuffers[outbuf.index].start, (void*)inBuffers[inbuf.index].start, inBuffers[inbuf.index].length);
	/* send output video buffer to output queue */
	if (-1 == xioctl(outfd, VIDIOC_QBUF, &outbuf))
		errno_exit("VIDIOC_QBUF error");

	/* obtain buffer for output VBI */
	CLEAR(outvbibuf);
	outvbibuf.type = outvbitype;
	if (-1 == xioctl(outvbifd, VIDIOC_DQBUF, &outvbibuf)) {
		switch (errno) {
		case EAGAIN:
			return 0;
		case EIO:
			/* Could ignore EIO, see spec. */
			/* fall through */
		default:
			errno_exit("VIDIOC_DQBUF error");
		}
	}
	assert(outvbibuf.index < outVBIBuffers.size());
	/* copy VBI data from input to output */
	memcpy((void*)outVBIBuffers[outvbibuf.index].start, (void*)inVBIBuffers[invbibuf.index].start, inVBIBuffers[invbibuf.index].length);
	/* send output VBI buffer to output queue */
	if (-1 == xioctl(outvbifd, VIDIOC_QBUF, &outvbibuf))
		errno_exit("VIDIOC_QBUF error");

	return 1;
}

static void mainloop(void)
{
	unsigned int count = 1000;
	while (count-- > 0) {
		for (;;) {
			fd_set fds;
			struct timeval tv;
			int r;
			FD_ZERO(&fds);
			FD_SET(infd, &fds);
			/* Timeout. */
			tv.tv_sec = 2;
			tv.tv_usec = 0;
			r = select(infd + 1, &fds, NULL, NULL, &tv);
			if (-1 == r) {
				if (EINTR == errno)
					continue;
				errno_exit("select error");
			}
			if (0 == r) {
				fprintf(stderr, "select timeout\n");
				exit(EXIT_FAILURE);
			}
			if (rewrite_frame())
				break;
			/* EAGAIN - continue select loop. */
		}
	}
}


static void uninit_devices(void)
{
	if (-1 == xioctl(outfd, VIDIOC_STREAMOFF, &outtype))
		errno_exit("Output streaming stop failed");
	if (-1 == xioctl(outvbifd, VIDIOC_STREAMOFF, &outvbitype))
		errno_exit("Output VBI streaming stop failed");
	if (-1 == xioctl(infd, VIDIOC_STREAMOFF, &intype))
		errno_exit("Input streaming stop failed");
	if (-1 == xioctl(invbifd, VIDIOC_STREAMOFF, &invbitype))
		errno_exit("Input VBI streaming stop failed");
	unsigned int i;
	for (i = 0; i < inBuffers.size(); ++i)
		if (-1 == munmap(inBuffers[i].start, inBuffers[i].length))
			errno_exit("munmap error");
	for (i = 0; i < outBuffers.size(); ++i)
		if (-1 == munmap(outBuffers[i].start, outBuffers[i].length))
			errno_exit("munmap error");
}

v4l2_format getFormat(int fd, int type)
{
	v4l2_format result;
	result.type = type;
	if (-1 == xioctl(fd, VIDIOC_G_FMT, &result))
		errno_exit("VIDIOC_G_FMT  error");
	return result;
}

std::vector<buffer> requestBuffers(int fd, int type)
{
	std::vector<buffer> result;

	v4l2_requestbuffers reqbuf;
	reqbuf.type = type;
	reqbuf.count = BUFF_NUMBERS;
	reqbuf.memory = V4L2_MEMORY_MMAP;

	if (-1 == xioctl(fd, VIDIOC_REQBUFS, &reqbuf))
		errno_exit("Buffers request failed");

	result.resize(reqbuf.count);

	for (int i = 0; i < reqbuf.count; i++) {
		struct v4l2_buffer buf;
		memset(&buf, 0, sizeof(buf));

		buf.type = reqbuf.type;
		buf.memory = V4L2_MEMORY_MMAP;
		buf.index = i;

		if (-1 == xioctl(fd, VIDIOC_QUERYBUF, &buf))
			errno_exit("VIDIOC_QUERYBUF error");

		result[i].index = i;
		result[i].length = buf.length;
		result[i].start = mmap(NULL, buf.length, PROT_READ | PROT_WRITE, MAP_SHARED, fd, buf.m.offset);

		if(result[i].start == MAP_FAILED)
			errno_exit("mmap error");;
	}

	return result;
}

static void init_devices(void) 
{
	struct v4l2_control ctl = { V4L2_CID_FORWARD_ENABLE_VANC, true };
	struct v4l2_buffer v4l2buf;
	if (-1 == xioctl(infd, VIDIOC_S_CTRL, &ctl))
		errno_exit("VANC input enabling failed");

	if (-1 == xioctl(outfd, VIDIOC_S_CTRL, &ctl))
		errno_exit("VANC output enabling failed");

	v4l2_format outFmt = getFormat(outfd, V4L2_BUF_TYPE_VIDEO_OUTPUT);
	v4l2_format vbiFmt = getFormat(outvbifd, V4L2_BUF_TYPE_VBI_OUTPUT);

	inBuffers = requestBuffers(infd, intype);
	outBuffers = requestBuffers(outfd, outtype);
	inVBIBuffers = requestBuffers(invbifd, invbitype);
	outVBIBuffers = requestBuffers(outvbifd, outvbitype);

	memset(&v4l2buf, 0, sizeof(v4l2buf));
	v4l2buf.type = intype;
	v4l2buf.memory = V4L2_MEMORY_MMAP;
	for (auto buf : inBuffers) {
		v4l2buf.index = buf.index;
		if (-1 == xioctl(infd, VIDIOC_QBUF, &v4l2buf))
			errno_exit("Input buffer queue failed");
	}

	memset(&v4l2buf, 0, sizeof(v4l2buf));
	v4l2buf.type = invbitype;
	v4l2buf.memory = V4L2_MEMORY_MMAP;
	for (auto buf : inVBIBuffers) {
		v4l2buf.index = buf.index;
		if (-1 == xioctl(invbifd, VIDIOC_QBUF, &v4l2buf))
			errno_exit("Input VBI buffer queue failed");
	}

	memset(&v4l2buf, 0, sizeof(v4l2buf));
	v4l2buf.type = outtype;
	v4l2buf.memory = V4L2_MEMORY_MMAP;
	for (auto buf : outBuffers) {
		v4l2buf.index = buf.index;
		v4l2buf.bytesused = outFmt.fmt.pix.sizeimage;
		memset(buf.start, 0, v4l2buf.bytesused);
		if (-1 == xioctl(outfd, VIDIOC_QBUF, &v4l2buf))
			errno_exit("Output queue failed");
	}

	memset(&v4l2buf, 0, sizeof(v4l2buf));
	v4l2buf.type = outvbitype;
	v4l2buf.memory = V4L2_MEMORY_MMAP;
	for (auto buf : outVBIBuffers) {
		v4l2buf.index = buf.index;
		v4l2buf.bytesused = vbiFmt.fmt.vbi.count[0] * vbiFmt.fmt.vbi.samples_per_line * 2;
		memset(buf.start, 0, v4l2buf.bytesused);  
		if (-1 == xioctl(outvbifd, VIDIOC_QBUF, &v4l2buf))
			errno_exit("Output VBI buffer queue failed");
	}

	if (-1 == xioctl(invbifd, VIDIOC_STREAMON, &invbitype))
		errno_exit("Input VBI streaming start failed");
	if (-1 == xioctl(infd, VIDIOC_STREAMON, &intype))
		errno_exit("Input streaming start failed");
	if (-1 == xioctl(outvbifd, VIDIOC_STREAMON, &outvbitype))
		errno_exit("Output VBI streaming start failed");
	if (-1 == xioctl(outfd, VIDIOC_STREAMON, &outtype))
		errno_exit("Output streaming start failed");

}

static void close_devices(void)
{
	free(buffer_in);
	snd_pcm_status_free(status_in);
	snd_pcm_hw_params_free(hw_params_in);
	snd_pcm_close(pcm_in);

	free(buffer_out);
	snd_pcm_status_free(status_out);
	snd_pcm_hw_params_free(hw_params_out);
	snd_pcm_close(pcm_out);

	if (-1 == close(infd))
		errno_exit("close infd error");
	infd = -1;
	if (-1 == close(invbifd))
		errno_exit("close invbifd error");
	invbifd = -1;
	if (-1 == close(outfd))
		errno_exit("close outfd error");
	outfd = -1;
	if (-1 == close(outvbifd))
		errno_exit("close outvbifd error");
	outvbifd = -1;
}

int open_dev(const std::string& name_dev, int reqCaps)
{
	struct stat st;
	if (-1 == stat(name_dev.c_str(), &st)) {
		fprintf(stderr, "Cannot identify '%s': %d, %s\n", name_dev.c_str(), errno, strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (!S_ISCHR(st.st_mode)) {
		fprintf(stderr, "%s is no device\n", name_dev.c_str());
		exit(EXIT_FAILURE);
	}
	int fd = open(name_dev.c_str(), O_RDWR /* required */ | O_NONBLOCK);
	if (-1 == fd) {
		fprintf(stderr, "Cannot open '%s': %d, %s\n", name_dev.c_str(), errno, strerror(errno));
		exit(EXIT_FAILURE);
	}
	struct v4l2_capability cap;
	if (-1 == xioctl(fd, VIDIOC_QUERYCAP, &cap)) {
		if (EINVAL == errno) {
			fprintf(stderr, "%s is no V4L2 device\n", name_dev.c_str());
			exit(EXIT_FAILURE);
		}
		else {
			errno_exit("VIDIOC_QUERYCAP error");
		}
	}
	if (!(cap.capabilities & reqCaps)) {
		fprintf(stderr, "%s.  Wrong device caps\n", name_dev.c_str());
		exit(EXIT_FAILURE);
	}
	if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
		fprintf(stderr, "%s does not support streaming i/o\n", name_dev.c_str());
		exit(EXIT_FAILURE);
	}
	return fd;
}

static void open_devices(void)
{
	infd = open_dev(dev_in, V4L2_CAP_VIDEO_CAPTURE);
	invbifd = open_dev(vbi_in, V4L2_CAP_VBI_CAPTURE);
	outfd = open_dev(dev_out, V4L2_CAP_VIDEO_OUTPUT);
	outvbifd = open_dev(vbi_out, V4L2_CAP_VBI_OUTPUT);
}


static void usage(FILE* fp, int argc, char** argv)
{
	fprintf(fp,
		"Usage: %s [options]\n\n"
		"Options:\n"
		"-d | --device number   Input Video device number [0]\n"
		"-o | --outdevice number   Output Video device number [2]\n"
		"-h | --help          Print this message\n"
		"",
		argv[0]);
}

static const char short_options[] = "d:o:hmru";

static const struct option
long_options[] = {
		{ "device",     required_argument,      NULL,           'd' },
		{ "outdevice",  required_argument,      NULL,           'o' },
		{ "help",       no_argument,            NULL,           'h' },
		{ 0, 0, 0, 0 }
};

static void command_prompt(int argc, char** argv)
{
	for (;;) {
		int index;
		int c;
		c = getopt_long(argc, argv, short_options, long_options, &index);
		if (-1 == c)
			break;
		switch (c) {
		case 0: /* getopt_long() flag */
			break;
		case 'd': {
			int k = atoi(optarg);
			switch (k) {
			case 1:
				dev_in = "/dev/video1";
				vbi_in = "/dev/vbi1";
				audio_in = "hw:0,1";
				break;
			case 0:
			default:
				dev_in = "/dev/video0";
				vbi_in = "/dev/vbi0";
				audio_in = "hw:0,0";
			}
		}
				break;
		case 'o': {
			int k = atoi(optarg);
			switch (k) {
			case 3:
				dev_out = "/dev/video3";
				vbi_out = "/dev/vbi3";
				audio_out = "hw:0,3";
				break;
			case 2:
			default:
				dev_out = "/dev/video2";
				vbi_out = "/dev/vbi2";
				audio_out = "hw:0,2";
			}
		}
				break;
		case 'h':
			usage(stdout, argc, argv);
			exit(EXIT_SUCCESS);
		default:
			usage(stderr, argc, argv);
			exit(EXIT_FAILURE);
		}
	}
}

int32_t* soundPrepare(snd_pcm_t* pcm, snd_pcm_hw_params_t** hw_params, snd_pcm_status_t** status)
{
	snd_pcm_sw_params_t *sw_params;
	unsigned int periods = 32;
	unsigned int channels = 4;
	int dir = 0;

	if(0 != snd_pcm_hw_params_malloc(hw_params))
		fprintf(stderr,"Cannot alloc hw params\n");

	if(0 != snd_pcm_status_malloc(status))
		fprintf(stderr,"Cannot alloc pcm status\n");

	if(0 != snd_pcm_hw_params_any(pcm, *hw_params))
		fprintf(stderr,"Cannot get any hw params\n");

	if (0 != snd_pcm_hw_params_set_access(pcm, *hw_params, SND_PCM_ACCESS_RW_INTERLEAVED))
		fprintf(stderr,"Cannot set pcm access mode\n");

	if (0 != snd_pcm_hw_params_set_period_size_near(pcm, *hw_params, &period, &dir))
		fprintf(stderr,"Cannot set pcm period size\n");

	dir = 0;
	if (0 != snd_pcm_hw_params_set_periods_near(pcm, *hw_params, &periods, &dir))
		fprintf(stderr,"Cannot set pcm periods\n");

	int ret = snd_pcm_hw_params_set_channels(pcm, *hw_params, channels);
	if (0 > ret)
		fprintf(stderr, "Cannot set %d channels.  %d, %s\n", channels, ret, strerror(ret));

	if (0 != snd_pcm_hw_params(pcm, *hw_params))
		fprintf(stderr,"Cannot set hw params\n");

	snd_pcm_sw_params_alloca(&sw_params);
	if (0 != snd_pcm_sw_params_current(pcm, sw_params))
		fprintf(stderr,"Cannot get sw params\n");

	if (0 != snd_pcm_sw_params_set_tstamp_mode(pcm, sw_params, SND_PCM_TSTAMP_ENABLE))
		fprintf(stderr,"Cannot enable pcm timestamping\n");

	if (0 != snd_pcm_sw_params(pcm, sw_params))
		fprintf(stderr,"Cannot set sw params\n");

	if (0 != snd_pcm_prepare(pcm))
		fprintf(stderr,"Cannot prepare pcm capture\n");
	
	snd_pcm_hw_params_get_format(*hw_params, &format);

	int32_t * buffer = (int32_t*)malloc(period * snd_pcm_format_width(format) / 8 * 8);
	return buffer;
}

void SoundPrepare(void)
{
	if(0 != snd_pcm_open(&pcm_in, audio_in.c_str(), SND_PCM_STREAM_CAPTURE, SND_PCM_NONBLOCK))
		fprintf(stderr,"Cannot open %s\n",audio_in.c_str());

	buffer_in = soundPrepare(pcm_in, &hw_params_in, &status_in);

	if(0 != snd_pcm_open(&pcm_out, audio_out.c_str(), SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK))
		fprintf(stderr,"Cannot open %s\n",audio_out.c_str());
	buffer_out = soundPrepare(pcm_out, &hw_params_out, &status_out);
}


int main(int argc, char** argv)
{
	command_prompt(argc, argv);

	open_devices();   // open video and VBI devices
	init_devices();   // init video and VBI devices
	SoundPrepare();   // open and prepare sound devices
	mainloop();
	uninit_devices();    // uninit video and VBI devices
	close_devices();     // close all devices
	return 0;
}
